from odoo import _, api, fields, models


class ResPartner(models.Model):
    """
    One2many relationship with CooperativeStatus should
    be replaced by inheritance.
    """

    _inherit = "res.partner"

    worker_store = fields.Boolean(default=False)
    is_worker = fields.Boolean(
        related="worker_store", string="Worker", readonly=False
    )
    can_shop = fields.Boolean(
        string="Is worker allowed to shop?",
        compute="_compute_can_shop",
        store=True,
    )
    cooperative_status_ids = fields.One2many(
        string="Cooperative Statuses",
        comodel_name="cooperative.status",
        inverse_name="cooperator_id",
        readonly=True,
    )
    super = fields.Boolean(
        related="cooperative_status_ids.super",
        string="Super Cooperative",
        readonly=True,
        store=True,
    )
    info_session = fields.Boolean(
        related="cooperative_status_ids.info_session",
        string="Information Session ?",
        readonly=True,
        store=True,
    )
    info_session_date = fields.Date(
        related="cooperative_status_ids.info_session_date",
        string="Information Session Date",
        readonly=True,
        store=True,
    )
    working_mode = fields.Selection(
        related="cooperative_status_ids.working_mode",
        readonly=True,
        store=True,
    )
    exempt_reason_id = fields.Many2one(
        related="cooperative_status_ids.exempt_reason_id",
        readonly=True,
        store=True,
    )
    state = fields.Selection(
        related="cooperative_status_ids.status", readonly=True, store=True
    )
    extension_start_time = fields.Date(
        related="cooperative_status_ids.extension_start_time",
        string="Extension Start Day",
        readonly=True,
        store=True,
    )
    subscribed_shift_ids = fields.Many2many("beesdoo.shift.template")

    @api.depends("cooperative_status_ids")
    def _compute_can_shop(self):
        """
        Shopping authorisation may vary on the can_shop status of the
        cooperative.status but also other parameters.
        Overwrite this function to change the default behavior.
        """
        for rec in self:
            if rec.cooperative_status_ids:
                rec.can_shop = rec.cooperative_status_ids.can_shop
            else:
                rec.can_shop = True

    @api.multi
    def coop_subscribe(self):
        return {
            "name": _("Subscribe Cooperator"),
            "type": "ir.actions.act_window",
            "view_type": "form",
            "view_mode": "form",
            "res_model": "beesdoo.shift.subscribe",
            "target": "new",
        }

    @api.multi
    def coop_unsubscribe(self):
        res = self.coop_subscribe()
        res["context"] = {"default_unsubscribed": True}
        return res

    @api.multi
    def manual_extension(self):
        return {
            "name": _("Manual Extension"),
            "type": "ir.actions.act_window",
            "view_type": "form",
            "view_mode": "form",
            "res_model": "beesdoo.shift.extension",
            "target": "new",
        }

    @api.multi
    def auto_extension(self):
        res = self.manual_extension()
        res["context"] = {"default_auto": True}
        res["name"] = _("Trigger Grace Delay")
        return res

    @api.multi
    def register_holiday(self):
        return {
            "name": _("Register Holiday"),
            "type": "ir.actions.act_window",
            "view_type": "form",
            "view_mode": "form",
            "res_model": "beesdoo.shift.holiday",
            "target": "new",
        }

    @api.multi
    def temporary_exempt(self):
        return {
            "name": _("Temporary Exemption"),
            "type": "ir.actions.act_window",
            "view_type": "form",
            "view_mode": "form",
            "res_model": "beesdoo.shift.temporary_exemption",
            "target": "new",
        }

    # TODO access right + vue on res.partner



from . import subscribe
from . import assign_super_coop
from . import batch_template
from . import extension
from . import holiday
from . import instanciate_planning
from . import temporary_exemption



# Copyright 2020 Coop IT Easy SCRL fs
#   Elouan Le Bars <elouan@coopiteasy.be>
# License AGPL-3.0 or later (https://www.gnu.org/licenses/agpl.html).


{
    "name": "Beescoop Shift Management",
    "summary": """Generate and manage shifts for cooperators.""",
    "author": "Thibault Francois, Elouan Le Bars, Coop IT Easy SCRLfs",
    "website": "https://github.com/beescoop/Obeesdoo",
    "category": "Cooperative management",
    "version": "12.0.1.0.1",
    "depends": ["mail"],
    "data": [
        "data/system_parameter.xml",
        "data/cron.xml",
        "data/mail_template.xml",
        "security/group.xml",
        "security/ir.model.access.csv",
        "views/task_template.xml",
        "views/task.xml",
        "views/planning.xml",
        "views/cooperative_status.xml",
        "views/exempt_reason.xml",
        "views/menu.xml",
        "wizard/instanciate_planning.xml",
        "wizard/batch_template.xml",
        "wizard/assign_super_coop.xml",
        "wizard/subscribe.xml",
        "wizard/extension.xml",
        "wizard/holiday.xml",
        "wizard/temporary_exemption.xml",
    ],
    "demo": ["demo/templates.xml", "demo/workers.xml"],
    "license": "AGPL-3",
}



from odoo import _, api, fields, models


class InstanciatePlanning(models.TransientModel):
    _name = "beesddoo.shift.generate_planning"
    _description = "beesddoo.shift.generate_planning"

    def _get_planning(self):
        return self._context.get("active_id")

    date_start = fields.Date(
        "First Day of planning (should be monday)", required=True
    )
    planning_id = fields.Many2one(
        "beesdoo.shift.planning", readonly=True, default=_get_planning
    )

    @api.multi
    def generate_task(self):
        self.ensure_one()
        self = self.with_context(
            visualize_date=self.date_start, tracking_disable=True
        )
        shifts = self.planning_id.task_template_ids._generate_task_day()
        return {
            "name": _("Generated Shift"),
            "type": "ir.actions.act_window",
            "view_type": "form",
            "view_mode": "kanban,calendar,tree,form,pivot",
            "res_model": "beesdoo.shift.shift",
            "target": "current",
            "domain": [("id", "in", shifts.ids)],
            "context": {"search_default_gb_day": 1},
        }



from odoo import _, api, fields, models
from odoo.exceptions import ValidationError


class TemporaryExemption(models.TransientModel):
    _name = "beesdoo.shift.temporary_exemption"
    _description = "beesdoo.shift.temporary_exemption"
    _inherit = "beesdoo.shift.action_mixin"

    temporary_exempt_reason_id = fields.Many2one(
        "cooperative.exempt.reason", "Exempt Reason", required=True
    )
    temporary_exempt_start_date = fields.Date(
        default=fields.Date.today, required=True
    )
    temporary_exempt_end_date = fields.Date(required=True)

    @api.multi
    def exempt(self):
        self = self._check()  # maybe a different group
        status_id = self.env["cooperative.status"].search(
            [("cooperator_id", "=", self.cooperator_id.id)]
        )
        if (
            status_id.temporary_exempt_end_date
            and status_id.temporary_exempt_end_date >= status_id.today
        ):
            raise ValidationError(
                _(
                    "You cannot encode new temporary exemption since the "
                    "previous one are not over yet "
                )
            )
        reason = self.temporary_exempt_reason_id
        status_id.sudo().write(
            {
                "temporary_exempt_start_date": self.temporary_exempt_start_date,
                "temporary_exempt_end_date": self.temporary_exempt_end_date,
                "temporary_exempt_reason_id": reason.id,
            }
        )
        self.env["beesdoo.shift.shift"].sudo().unsubscribe_from_today(
            [self.cooperator_id.id],
            today=self.temporary_exempt_start_date,
            end_date=self.temporary_exempt_end_date,
        )



import math
from datetime import datetime, time, timedelta

from pytz import UTC, timezone

from odoo import _, api, fields, models
from odoo.exceptions import UserError


def float_to_time(f):
    decimal, integer = math.modf(f)
    return "{}:{}".format(
        str(int(integer)).zfill(2), str(int(round(decimal * 60))).zfill(2)
    )


def floatime_to_hour_minute(f):
    decimal, integer = math.modf(f)
    return int(integer), int(round(decimal * 60))


def get_first_day_of_week():
    today = datetime.now()
    return (datetime.now() - timedelta(days=today.weekday())).date()


class TaskType(models.Model):
    _name = "beesdoo.shift.type"
    _description = "beesdoo.shift.type"

    name = fields.Char()
    description = fields.Text()
    active = fields.Boolean(default=True)


class DayNumber(models.Model):
    _name = "beesdoo.shift.daynumber"
    _description = "beesdoo.shift.daynumber"

    _order = "number asc"

    name = fields.Char()
    number = fields.Integer(
        "Day Number",
        help="From 1 to N, When you will instanciate your planning, Day 1 "
        "will be the start date of the instance, Day 2 the second, "
        "etc...",
    )
    active = fields.Boolean(default=True)


class Planning(models.Model):
    _name = "beesdoo.shift.planning"
    _description = "beesdoo.shift.planning"
    _order = "sequence asc"

    sequence = fields.Integer()
    name = fields.Char()
    task_template_ids = fields.One2many(
        "beesdoo.shift.template", "planning_id"
    )

    @api.model
    def _get_next_planning(self, sequence):
        next_planning = self.search([("sequence", ">", sequence)])
        if not next_planning:
            return self.search([])[0]
        return next_planning[0]

    @api.multi
    def _get_next_planning_date(self, date):
        self.ensure_one()
        nb_of_day = max(self.task_template_ids.mapped("day_nb_id.number"))
        return date + timedelta(days=nb_of_day)

    @api.model
    def _generate_next_planning(self):
        config = self.env["ir.config_parameter"].sudo()
        last_seq = int(config.get_param("last_planning_seq", 0))
        date = fields.Date.from_string(
            config.get_param("next_planning_date", 0)
        )

        planning = self._get_next_planning(last_seq)
        planning = planning.with_context(visualize_date=date)
        planning.task_template_ids._generate_task_day()

        next_date = planning._get_next_planning_date(date)
        config.set_param("last_planning_seq", planning.sequence)
        config.set_param("next_planning_date", next_date)


class TaskTemplate(models.Model):
    _name = "beesdoo.shift.template"
    _description = "beesdoo.shift.template"
    _order = "start_time"

    name = fields.Char(required=True)
    planning_id = fields.Many2one("beesdoo.shift.planning", required=True)
    day_nb_id = fields.Many2one(
        "beesdoo.shift.daynumber", string="Day", required=True
    )
    task_type_id = fields.Many2one("beesdoo.shift.type", string="Type")
    # FIXME removed because beesdoo.shift.sheet is from another module.
    # attendance_sheet_id = fields.Many2one('beesdoo.shift.sheet', string="Attendance Sheet")  # noqa
    start_time = fields.Float(required=True)
    end_time = fields.Float(required=True)
    super_coop_id = fields.Many2one(
        "res.users",
        string="Super Cooperative",
        domain=[("partner_id.super", "=", True)],
    )

    duration = fields.Float(help="Duration in Hour")
    worker_nb = fields.Integer(
        string="Number of worker",
        help="Max number of worker for this task",
        default=1,
    )
    worker_ids = fields.Many2many(
        "res.partner",
        string="Recurrent worker assigned",
        domain=[("is_worker", "=", True)],
    )
    remaining_worker = fields.Integer(
        compute="_compute_remaining", store=True, string="Remaining Place"
    )
    active = fields.Boolean(default=True)
    # For Kanban View Only
    color = fields.Integer("Color Index")
    worker_name = fields.Char(compute="_compute_worker_name")
    # For calendar View
    start_date = fields.Datetime(
        compute="_compute_fake_date", search="_search_dummy"
    )
    end_date = fields.Datetime(
        compute="_compute_fake_date", search="_search_dummy"
    )

    def _get_utc_date(self, day, hour, minute):
        """Combine day number, hours and minutes to save
        corresponding UTC datetime in database.
        """
        context_tz = timezone(self._context.get("tz") or self.env.user.tz)
        day_local_time = datetime.combine(day, time(hour=hour, minute=minute))
        day_local_time = context_tz.localize(day_local_time)
        day_utc_time = day_local_time.astimezone(UTC)
        # Return naïve datetime so as to be saved in database
        return day_utc_time.replace(tzinfo=None)

    @api.depends("start_time", "end_time")
    def _compute_fake_date(self):
        today = self._context.get("visualize_date", get_first_day_of_week())
        for rec in self:
            # Find the day of this task template 'rec'.
            day = today + timedelta(days=rec.day_nb_id.number - 1)
            # Compute the beginning and ending time according to the
            # context timezone.
            h_begin, m_begin = floatime_to_hour_minute(rec.start_time)
            h_end, m_end = floatime_to_hour_minute(rec.end_time)
            rec.start_date = self._get_utc_date(day, h_begin, m_begin)
            rec.end_date = self._get_utc_date(day, h_end, m_end)

    def _search_dummy(self, operator, value):
        return []

    @api.depends("worker_ids", "worker_nb")
    def _compute_remaining(self):
        for rec in self:
            rec.remaining_worker = rec.worker_nb - len(rec.worker_ids)

    @api.depends("worker_ids")
    def _compute_worker_name(self):
        for rec in self:
            rec.worker_name = ",".join(rec.worker_ids.mapped("display_name"))

    @api.constrains("worker_nb", "worker_ids")
    def _nb_worker_max(self):
        for rec in self:
            if len(rec.worker_ids) > rec.worker_nb:
                raise UserError(
                    _(
                        "You cannot assign more workers than the maximal "
                        "number defined on template. "
                    )
                )

    @api.onchange("start_time", "end_time")
    def _get_duration(self):
        if self.start_time and self.end_time:
            self.duration = self.end_time - self.start_time

    @api.onchange("duration")
    def _set_duration(self):
        if self.start_time:
            self.end_time = self.start_time + self.duration

    def _generate_task_day(self):
        tasks = self.env["beesdoo.shift.shift"]
        for rec in self:
            for i in range(0, rec.worker_nb):
                worker_id = (
                    rec.worker_ids[i] if len(rec.worker_ids) > i else False
                )
                # remove worker in holiday and temporary exempted
                if worker_id and worker_id.cooperative_status_ids:
                    status = worker_id.cooperative_status_ids[0]
                    if (
                        status.holiday_start_time
                        and status.holiday_end_time
                        and status.holiday_start_time <= rec.start_date.date()
                        and status.holiday_end_time >= rec.end_date.date()
                    ):
                        worker_id = False
                    if (
                        status.temporary_exempt_start_date
                        and status.temporary_exempt_end_date
                        and status.temporary_exempt_start_date
                        <= rec.start_date.date()
                        and status.temporary_exempt_end_date
                        >= rec.end_date.date()
                    ):
                        worker_id = False
                tasks |= tasks.create(
                    {
                        "name": "[%s] %s %s (%s - %s) [%s]"
                        % (
                            rec.start_date.date(),
                            rec.planning_id.name,
                            rec.day_nb_id.name,
                            float_to_time(rec.start_time),
                            float_to_time(rec.end_time),
                            i,
                        ),
                        "task_template_id": rec.id,
                        "task_type_id": rec.task_type_id.id,
                        "super_coop_id": rec.super_coop_id.id,
                        "worker_id": worker_id and worker_id.id or False,
                        "is_regular": True if worker_id else False,
                        "start_time": rec.start_date,
                        "end_time": rec.end_date,
                        "state": "open",
                    }
                )

        return tasks

    @api.onchange("worker_ids")
    def check_for_multiple_shifts(self):
        original_ids = {worker.id for worker in self._origin.worker_ids}

        warnings = []
        for worker in self.worker_ids:
            if worker.id not in original_ids:
                shifts = [
                    shift.name
                    for shift in worker.subscribed_shift_ids
                    if shift.id != self.id
                ]
                if shifts:
                    warnings.append(
                        worker.name
                        + _(" is already assigned to ")
                        + ", ".join(shifts)
                    )

        if warnings:
            return {
                "warning": {
                    "title": _("Warning"),
                    "message": "\n".join(warnings),
                }
            }



import logging
from datetime import timedelta

from odoo import _, api, fields, models
from odoo.exceptions import UserError, ValidationError

_logger = logging.getLogger(__name__)


def add_days_delta(date_from, days_delta):
    if not date_from:
        return date_from
    next_date = date_from + timedelta(days=days_delta)
    return next_date


class ExemptReason(models.Model):
    _name = "cooperative.exempt.reason"
    _description = "cooperative.exempt.reason"

    name = fields.Char(required=True)


class HistoryStatus(models.Model):
    _name = "cooperative.status.history"
    _description = "cooperative.status.history"

    _order = "create_date desc"

    status_id = fields.Many2one("cooperative.status")
    cooperator_id = fields.Many2one("res.partner")
    change = fields.Char()
    type = fields.Selection(
        [("status", "Status Change"), ("counter", "Counter Change")]
    )
    user_id = fields.Many2one("res.users", string="User")


class CooperativeStatus(models.Model):
    _name = "cooperative.status"
    _description = "cooperative.status"
    _rec_name = "cooperator_id"
    _order = "cooperator_id"
    _period = 28

    def _get_status(self):
        return [
            ("ok", "Up to Date"),
            ("holiday", "Holidays"),
            ("alert", "Alerte"),
            ("extension", "Extension"),
            ("suspended", "Suspended"),
            ("exempted", "Exempted"),
            ("unsubscribed", "Unsubscribed"),
            ("resigning", "Resigning"),
        ]

    today = fields.Date(
        help="Field that allow to compute field and store them even if they "
        "are based on the current date",
        default=fields.Date.today,
    )
    cooperator_id = fields.Many2one("res.partner")
    active = fields.Boolean(
        related="cooperator_id.active", store=True, index=True
    )
    info_session = fields.Boolean("Information Session ?")
    info_session_date = fields.Date("Information Session Date")
    super = fields.Boolean("Super Cooperative")
    sr = fields.Integer("Regular shifts counter", default=0)
    sc = fields.Integer("Compensation shifts counter", default=0)
    time_extension = fields.Integer(
        "Extension Days NB",
        default=0,
        help="Addtional days to the automatic extension, 5 mean that you have "
        "a total of 15 extension days of default one is set to 10",
    )
    holiday_start_time = fields.Date("Holidays Start Day")
    holiday_end_time = fields.Date("Holidays End Day")
    alert_start_time = fields.Date("Alert Start Day")
    extension_start_time = fields.Date("Extension Start Day")
    working_mode = fields.Selection(
        [
            ("regular", "Regular worker"),
            ("irregular", "Irregular worker"),
            ("exempt", "Exempted"),
        ],
        string="Working mode",
    )
    exempt_reason_id = fields.Many2one(
        comodel_name="cooperative.exempt.reason", string="Exempt Reason"
    )
    status = fields.Selection(
        selection=_get_status,
        compute="_compute_status",
        string="Cooperative Status",
        store=True,
    )
    can_shop = fields.Boolean(compute="_compute_can_shop", store=True)
    history_ids = fields.One2many(
        "cooperative.status.history", "status_id", readonly=True
    )
    unsubscribed = fields.Boolean(default=False, help="Manually unsubscribed")
    resigning = fields.Boolean(
        default=False, help="Want to leave the beescoop"
    )

    # Specific to irregular
    irregular_start_date = fields.Date()  # TODO migration script
    irregular_absence_date = fields.Date()
    irregular_absence_counter = (
        fields.Integer()
    )  # TODO unsubscribe when reach -2
    future_alert_date = fields.Date(compute="_compute_future_alert_date")
    next_countdown_date = fields.Date(compute="_compute_next_countdown_date")

    temporary_exempt_reason_id = fields.Many2one(
        comodel_name="cooperative.exempt.reason",
        string="Temporary Exempt Reason",
    )
    temporary_exempt_start_date = fields.Date()
    temporary_exempt_end_date = fields.Date()

    @api.depends("status")
    def _compute_can_shop(self):
        for rec in self:
            rec.can_shop = rec.status in self._can_shop_status()

    @api.depends(
        "today",
        "sr",
        "sc",
        "holiday_end_time",
        "holiday_start_time",
        "time_extension",
        "alert_start_time",
        "extension_start_time",
        "unsubscribed",
        "irregular_absence_date",
        "irregular_absence_counter",
        "temporary_exempt_start_date",
        "temporary_exempt_end_date",
        "resigning",
        "cooperator_id.subscribed_shift_ids",
    )
    def _compute_status(self):
        update = int(
            self.env["ir.config_parameter"]
            .sudo()
            .get_param("always_update", False)
        )
        for rec in self:
            if update or not rec.today:
                rec.status = "ok"
                continue
            if rec.resigning:
                rec.status = "resigning"
                continue

            if rec.working_mode == "regular":
                rec.status = rec._get_regular_status()
            elif rec.working_mode == "irregular":
                rec.status = rec._get_irregular_status()
            elif rec.working_mode == "exempt":
                rec.status = "ok"

    _sql_constraints = [
        (
            "cooperator_uniq",
            "unique (cooperator_id)",
            _("You can only set one cooperator status per cooperator"),
        )
    ]

    @api.constrains("working_mode", "irregular_start_date")
    def _constrains_irregular_start_date(self):
        if self.working_mode == "irregular" and not self.irregular_start_date:
            raise UserError(
                _("Irregular workers must have an irregular start date.")
            )

    @api.multi
    def write(self, vals):
        """
            Overwrite write to historize the change
        """
        for field in [
            "sr",
            "sc",
            "time_extension",
            "extension_start_time",
            "alert_start_time",
            "unsubscribed",
        ]:
            if field not in vals:
                continue
            for rec in self:
                data = {
                    "status_id": rec.id,
                    "cooperator_id": rec.cooperator_id.id,
                    "type": "counter",
                    "user_id": self.env.context.get("real_uid", self.env.uid),
                }
                if vals.get(field, rec[field]) != rec[field]:
                    data["change"] = "{}: {} -> {}".format(
                        field.upper(), rec[field], vals.get(field)
                    )
                    self.env["cooperative.status.history"].sudo().create(data)
        return super(CooperativeStatus, self).write(vals)

    @api.multi
    def _write(self, vals):
        """
            Overwrite write to historize the change of status
            and make action on status change
        """
        if "status" in vals:
            self._cr.execute(
                'select id, status, sr, sc from "%s" where id in %%s'
                % self._table,
                (self._ids,),
            )
            result = self._cr.dictfetchall()
            old_status_per_id = {r["id"]: r for r in result}
            for rec in self:
                if old_status_per_id[rec.id]["status"] != vals["status"]:
                    data = {
                        "status_id": rec.id,
                        "cooperator_id": rec.cooperator_id.id,
                        "type": "status",
                        "change": "STATUS: %s -> %s"
                        % (
                            old_status_per_id[rec.id]["status"],
                            vals["status"],
                        ),
                        "user_id": self.env.context.get(
                            "real_uid", self.env.uid
                        ),
                    }
                    self.env["cooperative.status.history"].sudo().create(data)
                    rec._state_change(vals["status"])
        return super(CooperativeStatus, self)._write(vals)

    def get_status_value(self):
        """
        Workararound to get translated selection value instead of key in mail
        template.
        """
        state_list = (
            self.env["cooperative.status"]
            ._fields["status"]
            ._description_selection(self.env)
        )
        return dict(state_list)[self.status]

    @api.model
    def _set_today(self):
        """
            Method call by the cron to update store value base on the date
        """
        self.search([]).write({"today": fields.Date.today()})

    @api.model
    def _cron_compute_counter_irregular(self, today=False):
        """
            Journal ensure that a irregular worker will be only check
            once per day
        """
        today = today or fields.Date.today()
        journal = self.env["beesdoo.shift.journal"].search(
            [("date", "=", today)]
        )
        if not journal:
            journal = self.env["beesdoo.shift.journal"].create({"date": today})

        domain = self._get_irregular_worker_domain(today=today)
        irregular = self.search(domain)
        for status in irregular:
            delta = (today - status.irregular_start_date).days
            if (
                delta
                and delta % self._period == 0
                and status not in journal.line_ids
            ):
                status._change_irregular_counter()
                journal.line_ids |= status

    @api.multi
    def clear_history(self):
        self.ensure_one()
        self.history_ids.unlink()

    ########################################################
    #                   Method to override                 #
    #           To define the behavior of the status       #
    #                                                      #
    #       By default: everyone is always up to date      #
    ########################################################

    ##############################
    #   Computed field section   #
    ##############################
    @api.depends("today")
    def _compute_future_alert_date(self):
        """
            Compute date until the worker is up to date
            for irregular worker
        """
        for rec in self:
            rec.future_alert_date = False

    @api.depends("today")
    def _compute_next_countdown_date(self):
        """
        Compute the following countdown date. This date is the date when
        the worker will see his counter changed due to the cron. This
        date is like the birthday date of the worker that occurred each
        _period.
        """
        for rec in self:
            rec.next_countdown_date = False

    def _can_shop_status(self):
        """
            return the list of status that give access
            to active cooperator privilege
        """
        return ["ok", "alert", "extension", "exempted"]

    #####################################
    #   Status Change implementation    #
    #####################################

    def _get_regular_status(self):
        """
            Return the value of the status
            for the regular worker
        """
        return "ok"

    def _get_irregular_status(self):
        """
            Return the value of the status
            for the irregular worker
        """
        return "ok"

    def _state_change(self, new_state):
        """
            Hook to watch change in the state
        """
        pass

    def _change_counter(self, data):
        """
            Call when a shift state is changed
            use data generated by _get_counter_date_state_change
        """
        pass

    ###############################################
    #        Irregular Cron implementation        #
    ###############################################

    def _get_irregular_worker_domain(self):
        """
            return the domain the give the list
            of valid irregular worker that should
            get their counter changed by the cron
        """
        return [(0, "=", 1)]

    def _change_irregular_counter(self):
        """
            Define how the counter will change
            for the irregular worker
            where today - start_date is a multiple of the period
            by default 28 days
        """
        pass


class ShiftCronJournal(models.Model):
    _name = "beesdoo.shift.journal"
    _description = "beesdoo.shift.journal"
    _order = "date desc"
    _rec_name = "date"

    date = fields.Date()
    line_ids = fields.Many2many("cooperative.status")

    _sql_constraints = [
        (
            "one_entry_per_day",
            "unique (date)",
            _("You can only create one journal per day"),
        )
    ]

    @api.multi
    def run(self):
        self.ensure_one()
        if not self.user_has_groups("beesdoo_shift.group_cooperative_admin"):
            raise ValidationError(
                _("You don't have the access to perform this action")
            )
        self.sudo().env["cooperative.status"]._cron_compute_counter_irregular(
            today=self.date
        )



from odoo import _, api, fields, models
from odoo.exceptions import ValidationError


class Subscribe(models.TransientModel):
    _name = "beesdoo.shift.holiday"
    _description = "beesdoo.shift.holiday"
    _inherit = "beesdoo.shift.action_mixin"

    holiday_start_day = fields.Date(
        string="Start date for the holiday", default=fields.Date.today
    )
    holiday_end_day = fields.Date(string="End date for the holiday (included)")

    @api.multi
    def holidays(self):
        self = self._check()  # maybe a different group
        status_id = self.env["cooperative.status"].search(
            [("cooperator_id", "=", self.cooperator_id.id)]
        )
        if (
            status_id.holiday_end_time
            and status_id.holiday_end_time >= status_id.today
        ):
            raise ValidationError(
                _(
                    "You cannot encode new holidays since the previous "
                    "holidays encoded are not over yet "
                )
            )
        status_id.sudo().write(
            {
                "holiday_start_time": self.holiday_start_day,
                "holiday_end_time": self.holiday_end_day,
            }
        )
        self.env["beesdoo.shift.shift"].sudo().unsubscribe_from_today(
            [self.cooperator_id.id],
            today=self.holiday_start_day,
            end_date=self.holiday_end_day,
        )



from odoo import _, api, fields, models
from odoo.exceptions import UserError


class Subscribe(models.TransientModel):
    _name = "beesdoo.shift.extension"
    _description = "beesdoo.shift.extension"
    _inherit = "beesdoo.shift.action_mixin"

    def _get_default_extension_delay(self):
        return int(
            self.env["ir.config_parameter"]
            .sudo()
            .get_param("default_extension_delay", 28)
        )

    extension_start_date = fields.Date(
        string="Start date for the extension",
        default=fields.Date.today,
        readonly=True,
    )
    auto = fields.Boolean("Auto Extension", default=False)
    extension_days = fields.Integer(default=_get_default_extension_delay)

    @api.multi
    def auto_ext(self):
        self = self._check(group="beesdoo_shift.group_shift_attendance")
        status_id = self.env["cooperative.status"].search(
            [("cooperator_id", "=", self.cooperator_id.id)]
        )
        status_id.sudo().write(
            {"extension_start_time": self.extension_start_date}
        )

    @api.multi
    def extension(self):
        self = self._check()  # maybe a different group
        grace_delay = int(
            self.env["ir.config_parameter"]
            .sudo()
            .get_param("default_grace_delay", 10)
        )
        status_id = self.env["cooperative.status"].search(
            [("cooperator_id", "=", self.cooperator_id.id)]
        )
        if not status_id.extension_start_time:
            raise UserError(
                _(
                    "You should not make a manual extension when the grace "
                    "delay has not been triggered yet "
                )
            )
        today_delay = (
            status_id.today - status_id.extension_start_time
        ).days - grace_delay
        if today_delay < 0:
            raise UserError(
                _(
                    "You should not start a manual extension during the grace "
                    "delay "
                )
            )
        status_id.sudo().write(
            {"time_extension": self.extension_days + today_delay}
        )



from odoo import api, fields, models


class AssignSuperCoop(models.TransientModel):
    _name = "beesddoo.shift.assign_super_coop"
    _description = "beesddoo.shift.assign_super_coop"

    super_coop_id = fields.Many2one(
        "res.users",
        "New Super Cooperative",
        required=True,
        domain=[("super", "=", True)],
    )
    shift_ids = fields.Many2many(
        "beesdoo.shift.shift",
        readonly=True,
        default=lambda self: self._context.get("active_ids"),
    )

    @api.multi
    def write_super_coop(self):
        self.ensure_one()
        self.shift_ids.write({"super_coop_id": self.super_coop_id.id})



from odoo import _, api, fields, models
from odoo.exceptions import UserError


class StatusActionMixin(models.AbstractModel):
    _name = "beesdoo.shift.action_mixin"
    _description = "beesdoo.shift.action_mixin"

    cooperator_id = fields.Many2one(
        "res.partner",
        default=lambda self: self.env["res.partner"].browse(
            self._context.get("active_id")
        ),
        required=True,
    )

    def _check(self, group="beesdoo_shift.group_shift_management"):
        self.ensure_one()
        if not self.env.user.has_group(group):
            raise UserError(
                _("You don't have the required access for this operation.")
            )
        if (
            self.cooperator_id == self.env.user.partner_id
            and not self.env.user.has_group(
                "beesdoo_shift.group_cooperative_admin"
            )
        ):
            raise UserError(_("You cannot perform this operation on yourself"))
        return self.with_context(real_uid=self._uid)


class Subscribe(models.TransientModel):
    _name = "beesdoo.shift.subscribe"
    _description = "beesdoo.shift.subscribe"
    _inherit = "beesdoo.shift.action_mixin"

    def _get_date(self):
        date = (
            self.env["res.partner"]
            .browse(self._context.get("active_id"))
            .info_session_date
        )
        if not date:
            return fields.Date.today()
        else:
            return date

    def _get_info_session_date(self):
        date = (
            self.env["res.partner"]
            .browse(self._context.get("active_id"))
            .info_session_date
        )
        if date and self._get_info_session_followed():
            return date
        else:
            return False

    def _get_info_session_followed(self):
        session_followed = (
            self.env["res.partner"]
            .browse(self._context.get("active_id"))
            .info_session
        )
        return session_followed

    def _get_shift(self):
        shifts = (
            self.env["res.partner"]
            .browse(self._context.get("active_id"))
            .subscribed_shift_ids
        )
        if shifts:
            return shifts[0]
        return

    def _get_nb_shifts(self):
        return len(
            self.env["res.partner"]
            .browse(self._context.get("active_id"))
            .subscribed_shift_ids
        )

    def _get_super(self):
        return (
            self.env["res.partner"]
            .browse(self._context.get("active_id"))
            .super
        )

    def _get_mode(self):
        return (
            self.env["res.partner"]
            .browse(self._context.get("active_id"))
            .working_mode
        )

    def _get_reset_counter_default(self):
        partner = self.env["res.partner"].browse(
            self._context.get("active_id")
        )
        return (
            partner.state == "unsubscribed"
            and partner.working_mode == "regular"
        )

    info_session = fields.Boolean(
        string="Followed an information session",
        default=_get_info_session_followed,
    )
    info_session_date = fields.Date(
        string="Date of information session", default=_get_info_session_date
    )
    super = fields.Boolean(string="Super Cooperator", default=_get_super)
    working_mode = fields.Selection(
        [
            ("regular", "Regular worker"),
            ("irregular", "Irregular worker"),
            ("exempt", "Exempted"),
        ],
        default=_get_mode,
    )
    exempt_reason_id = fields.Many2one(
        "cooperative.exempt.reason", "Exempt Reason"
    )
    shift_id = fields.Many2one("beesdoo.shift.template", default=_get_shift)
    nb_shifts = fields.Integer(
        string="Number of shifts", default=_get_nb_shifts
    )
    reset_counter = fields.Boolean(default=_get_reset_counter_default)
    reset_compensation_counter = fields.Boolean(default=False)
    unsubscribed = fields.Boolean(
        default=False,
        string="Are you sure to remove this cooperator from his subscribed shift ?",
    )
    irregular_start_date = fields.Date(
        string="Start Date", default=fields.Date.today
    )
    resigning = fields.Boolean(
        default=False, help="Want to leave the beescoop"
    )

    @api.multi
    def unsubscribe(self):
        self = self._check()
        if not self.unsubscribed:
            return

        status_id = self.env["cooperative.status"].search(
            [("cooperator_id", "=", self.cooperator_id.id)]
        )
        data = {
            "unsubscribed": True,
            "cooperator_id": self.cooperator_id.id,
            "resigning": self.resigning,
        }
        if status_id:
            status_id.sudo().write(data)
        else:
            self.env["cooperative.status"].sudo().create(data)

    @api.multi
    def subscribe(self):
        self = self._check()
        if self.shift_id and self.shift_id.remaining_worker <= 0:
            raise UserError(_("There is no remaining space for this shift"))
        if self.shift_id:
            # Remove existing shift then subscribe to the new shift
            self.cooperator_id.sudo().write(
                {"subscribed_shift_ids": [(6, 0, [self.shift_id.id])]}
            )
        if self.working_mode != "regular":
            # Remove existing shift then subscribe to the new shift
            self.cooperator_id.sudo().write({"subscribed_shift_ids": [(5,)]})

        data = {
            "info_session": self.info_session,
            "info_session_date": self.info_session_date,
            "working_mode": self.working_mode,
            "exempt_reason_id": self.exempt_reason_id.id,
            "super": self.super,
            "cooperator_id": self.cooperator_id.id,
            "unsubscribed": False,
            "irregular_start_date": self.irregular_start_date,
            "irregular_absence_date": False,
            "irregular_absence_counter": 0,
        }
        if self.reset_counter:
            data["sr"] = 0
            data["extension_start_time"] = False
            data["alert_start_time"] = False
            data["time_extension"] = 0
        if self.reset_compensation_counter:
            data["sc"] = 0

        coop_status_obj = self.env["cooperative.status"]
        status_id = coop_status_obj.search(
            [("cooperator_id", "=", self.cooperator_id.id)]
        )
        if status_id:
            status_id.sudo().write(data)
        else:
            status_id = coop_status_obj.sudo().create(data)
            # Normally the write method is not necessary here.
            # But it does not work without it. You have to make 2 registration
            # to a shift to keep information like "Worker mode, session info
            # ,...
            status_id.sudo().write(data)
        return True



import json
from datetime import datetime, time, timedelta

from odoo import _, api, fields, models
from odoo.exceptions import UserError, ValidationError


class Task(models.Model):
    _name = "beesdoo.shift.shift"

    _inherit = ["mail.thread"]

    _order = "start_time asc"

    ##################################
    # Method to override             #
    # to have different state        #
    # on the shift                   #
    ##################################
    def _get_selection_status(self):
        return [
            ("open", "Confirmed"),
            ("done", "Attended"),
            ("absent", "Absent"),
            ("excused", "Excused"),
            ("cancel", "Cancelled"),
        ]

    def _get_color_mapping(self, state):
        return {
            "draft": 0,
            "open": 1,
            "done": 5,
            "absent": 2,
            "excused": 3,
            "cancel": 9,
        }[state]

    def _get_final_state(self):
        return ["done", "absent", "excused"]

    name = fields.Char(track_visibility="always")
    task_template_id = fields.Many2one("beesdoo.shift.template")
    planning_id = fields.Many2one(
        related="task_template_id.planning_id", store=True
    )
    task_type_id = fields.Many2one("beesdoo.shift.type", string="Task Type")
    worker_id = fields.Many2one(
        "res.partner",
        track_visibility="onchange",
        domain=[
            ("is_worker", "=", True),
            ("working_mode", "in", ("regular", "irregular")),
            ("state", "not in", ("unsubscribed", "resigning")),
        ],
    )
    start_time = fields.Datetime(
        track_visibility="always", index=True, required=True
    )
    end_time = fields.Datetime(track_visibility="always", required=True)
    state = fields.Selection(
        selection=_get_selection_status,
        default="open",
        required=True,
        track_visibility="onchange",
        group_expand="_expand_states",
    )
    color = fields.Integer(compute="_compute_color")
    super_coop_id = fields.Many2one(
        "res.users",
        string="Super Cooperative",
        domain=[("partner_id.super", "=", True)],
        track_visibility="onchange",
    )
    is_regular = fields.Boolean(default=False, string="Regular shift")
    is_compensation = fields.Boolean(
        default=False, string="Compensation shift"
    )
    replaced_id = fields.Many2one(
        "res.partner",
        track_visibility="onchange",
        domain=[
            ("eater", "=", "worker_eater"),
            ("working_mode", "=", "regular"),
            ("state", "not in", ("unsubscribed", "resigning")),
        ],
    )
    revert_info = fields.Text(copy=False)
    working_mode = fields.Selection(related="worker_id.working_mode")

    def _expand_states(self, states, domain, order):
        return [key for key, val in self._fields["state"].selection(self)]

    @api.depends("state")
    def _compute_color(self):
        for rec in self:
            rec.color = self._get_color_mapping(rec.state)

    def _compensation_validation(self, task):
        """
        Raise a validation error if the fields is_regular and
        is_compensation are not properly set.
        """
        if task.is_regular == task.is_compensation or not (
            task.is_regular or task.is_compensation
        ):
            raise ValidationError(
                _(
                    "You must choose between Regular Shift or "
                    "Compensation Shift."
                )
            )

    @api.constrains("state")
    def _lock_future_task(self):
        if datetime.now() < self.start_time:
            if self.state in self._get_final_state():
                raise UserError(
                    _(
                        "Shift state of a future shift "
                        "can't be set to 'present' or 'absent'."
                    )
                )

    @api.constrains("is_regular", "is_compensation")
    def _check_compensation(self):
        for task in self:
            if task.working_mode == "regular":
                self._compensation_validation(task)

    @api.constrains("worker_id")
    def _check_worker_id(self):
        """
        When worker_id changes we need to check whether is_regular
        and is_compensation are set correctly.
        When worker_id is set to a worker that doesn't need field
        is_regular and is_compensation, these two fields are set to
        False.
        """
        for task in self:
            if task.working_mode == "regular":
                self._compensation_validation(task)
            else:
                task.write({"is_regular": False, "is_compensation": False})
            if task.worker_id:
                if task.worker_id == task.replaced_id:
                    raise UserError(_("A worker cannot replace himself."))

    def message_auto_subscribe(self, updated_fields, values=None):
        self._add_follower(values)
        return super(Task, self).message_auto_subscribe(
            updated_fields, values=values
        )

    def _add_follower(self, vals):
        if vals.get("worker_id"):
            worker = self.env["res.partner"].browse(vals["worker_id"])
            self.message_subscribe(partner_ids=worker.ids)

    # TODO button to replaced someone
    @api.model
    def unsubscribe_from_today(
        self, worker_ids, today=None, end_date=None, now=None
    ):
        """
        Unsubscribe workers from *worker_ids* from all shift that start
          *today* and later.
        If *end_date* is given, unsubscribe workers from shift between *today*
          and *end_date*.
        If *now* is given workers are unsubscribed from all shifts starting
           *now* and later.
        If *now* is given, *end_date* is not taken into account.

        :type today: date
        :type end_date: date
        :type now: datetime
        """
        if now:
            if not isinstance(now, datetime):
                raise UserError(_("'Now' must be a datetime."))
            date_domain = [("start_time", ">", now)]
        else:
            today = today or fields.Date.today()
            today = datetime.combine(today, time())
            date_domain = [("start_time", ">", today)]
            if end_date:
                end_date = datetime.combine(
                    end_date, time(hour=23, minute=59, second=59)
                )
                date_domain.append(("end_time", "<=", end_date))

        to_unsubscribe = self.search(
            [("worker_id", "in", worker_ids)] + date_domain
        )
        to_unsubscribe.write({"worker_id": False})

        # Remove worker, replaced_id and regular
        to_unsubscribe_replace = self.search(
            [("replaced_id", "in", worker_ids)] + date_domain
        )
        to_unsubscribe_replace.write(
            {"worker_id": False, "replaced_id": False}
        )

        # If worker is Super cooperator, remove it from planning
        super_coop_ids = (
            self.env["res.users"]
            .search([("partner_id", "in", worker_ids), ("super", "=", True)])
            .ids
        )

        if super_coop_ids:
            to_unsubscribe_super_coop = self.search(
                [("super_coop_id", "in", super_coop_ids)] + date_domain
            )
            to_unsubscribe_super_coop.write({"super_coop_id": False})

    @api.multi
    def write(self, vals):
        """
            Overwrite write to track state change
            If worker is changer:
               Revert for the current worker
               Change the worker info
               Compute state change for the new worker
        """
        if "worker_id" in vals:
            for rec in self:
                if rec.worker_id.id != vals["worker_id"]:
                    rec._revert()
                    # To satisfy the constrains on worker_id, it must be
                    # accompanied by the change in is_regular and
                    # is_compensation field.
                    super(Task, rec).write(
                        {
                            "worker_id": vals["worker_id"],
                            "is_regular": vals.get(
                                "is_regular", rec.is_regular
                            ),
                            "is_compensation": vals.get(
                                "is_compensation", rec.is_compensation
                            ),
                        }
                    )
                    rec._update_state(rec.state)
        if "state" in vals:
            for rec in self:
                if vals["state"] != rec.state:
                    rec._update_state(vals["state"])
        return super(Task, self).write(vals)

    def _set_revert_info(self, data, status):
        data_new = {
            "status_id": status.id,
            "data": {
                k: data.get(k, 0) * -1
                for k in ["sr", "sc", "irregular_absence_counter"]
            },
        }
        if data.get("irregular_absence_date"):
            data_new["data"]["irregular_absence_date"] = False

        self.write({"revert_info": json.dumps(data_new)})

    def _revert(self):
        if not self.revert_info:
            return
        data = json.loads(self.revert_info)
        self.env["cooperative.status"].browse(
            data["status_id"]
        ).sudo()._change_counter(data["data"])
        self.revert_info = False

    def _update_state(self, new_state):
        self.ensure_one()
        self._revert()

        if (
            not (self.worker_id or self.replaced_id)
            and new_state in self._get_final_state()
        ):
            raise UserError(
                _(
                    "You cannot change to the status %s if no worker is "
                    "defined for the shift "
                )
                % new_state
            )

        always_update = int(
            self.env["ir.config_parameter"]
            .sudo()
            .get_param("always_update", False)
        )
        if always_update or not (self.worker_id or self.replaced_id):
            return

        if not (self.worker_id.working_mode in ["regular", "irregular"]):
            raise UserError(
                _(
                    "Working mode is not properly defined. Please check if "
                    "the worker is subscribed "
                )
            )

        data, status = self._get_counter_date_state_change(new_state)
        if status:
            status.sudo()._change_counter(data)
            self._set_revert_info(data, status)

    @api.model
    def _cron_send_weekly_emails(self):
        """
        Send a summary email for all workers
        if they have a shift planned during the week.
        """
        tasks = self.env["beesdoo.shift.shift"]
        shift_summary_mail_template = self.env.ref(
            "beesdoo_shift.email_template_shift_summary", False
        )

        start_time = datetime.now() + timedelta(days=1)
        end_time = datetime.now() + timedelta(days=7)

        confirmed_tasks = tasks.search(
            [
                ("start_time", ">", start_time),
                ("start_time", "<", end_time),
                ("worker_id", "!=", False),
                ("state", "=", "open"),
            ]
        )

        for rec in confirmed_tasks:
            shift_summary_mail_template.send_mail(rec.id, True)

    ########################################################
    #                   Method to override                 #
    #           To define the behavior of the status       #
    #                                                      #
    #       By default: everyone is always up to date      #
    ########################################################

    def _get_counter_date_state_change(self, new_state):
        """
        Return the cooperator_status of the cooperator that need to be
        change and data that need to be change. It does not perform the
        change directly. The cooperator_status will be changed by the
        _change_counter function.

        Check has been done to ensure that worker is legitimate.
        """
        data = {}
        status = None
        return data, status



"""
Created on 2 janv. 2017

@author: Thibault Francois
"""

from odoo import _, api, fields, models


class GenerateShiftTemplate(models.TransientModel):
    _name = "beesddoo.shift.generate_shift_template"
    _description = "beesddoo.shift.generate_shift_template"

    day_ids = fields.Many2many(
        "beesdoo.shift.daynumber",
        relation="template_gen_day_number_rel",
        column1="wizard_id",
        column2="day_id",
    )
    planning_id = fields.Many2one("beesdoo.shift.planning", required=True)
    type_id = fields.Many2one(
        "beesdoo.shift.type",
        default=lambda self: self._context.get("active_id"),
    )
    line_ids = fields.One2many(
        "beesddoo.shift.generate_shift_template.line", "wizard_id"
    )

    @api.multi
    def generate(self):
        self.ensure_one()
        ids = []
        for day in self.day_ids:
            for line in self.line_ids:
                shift_template_data = {
                    "name": "%s" % self.type_id.name,
                    "planning_id": self.planning_id.id,
                    "task_type_id": self.type_id.id,
                    "day_nb_id": day.id,
                    "start_time": line.start_time,
                    "end_time": line.end_time,
                    "duration": line.end_time - line.start_time,
                    "worker_nb": line.worker_nb,
                }
                new_rec = self.env["beesdoo.shift.template"].create(
                    shift_template_data
                )
                ids.append(new_rec.id)
        return {
            "name": _("Generated Shift Template"),
            "type": "ir.actions.act_window",
            "view_type": "form",
            "view_mode": "kanban,tree,form",
            "res_model": "beesdoo.shift.template",
            "target": "current",
            "domain": [("id", "in", ids)],
            "context": {"group_by": "day_nb_id"},
        }


class GenerateShiftTemplateLine(models.TransientModel):
    _name = "beesddoo.shift.generate_shift_template.line"
    _description = "beesddoo.shift.generate_shift_template.line"

    wizard_id = fields.Many2one("beesddoo.shift.generate_shift_template")
    start_time = fields.Float(required=True)
    end_time = fields.Float(required=True)
    worker_nb = fields.Integer(default=1)
